/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002, 2004, 2005                                  */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

package com.ibm.staf.service.eventmanager;

import com.ibm.staf.*;
import com.ibm.staf.service.*;
import java.util.*;
import java.io.*;
import org.python.util.PythonInterpreter;
import org.python.core.*;
import java.util.Properties;

public class EventManagerService implements STAFServiceInterfaceLevel30, Runnable
{
    private STAFHandle fHandle;
    private final String kVersion = "3.0.1";

    // Version of STAF (or later) required for this service
    private final String kRequiredSTAFVersion = "3.0.0";

    private STAFCommandParser fParmsParser = new STAFCommandParser();
    private STAFCommandParser fRegisterParser;
    private STAFCommandParser fListParser;
    private STAFCommandParser fVersionParser;
    private STAFMapClassDefinition fEventManagerIDMapClass;
    private STAFMapClassDefinition fEMIDMapClass;
    private STAFMapClassDefinition fSettingsMapClass;
    private Hashtable fRegTable;
    private static boolean DEBUG = true;
    private Thread fQueueThread;
    private STAFCommandParser fUnregisterParser;
    private int fEventManagerID = 1;
    private String fLineSep;
    private String fEventServiceMachine = "";
    private String fEventServiceName = "";
    private String fLocalMachineName = "";
    private String fServiceName = "eventmanager";
    private String fJythonVersion = "";
    private PythonInterpreter fPythonInterpreter = new PythonInterpreter();
    private String fHashtableFileDirectory;
    private String fHashtableFileName;
    private HashMap fSubmittedRequests = new HashMap();
    static final String sQueueTypeEnd = "STAF/Service/EventManager/End";

    public EventManagerService()
    {
    }

    public STAFResult init(STAFServiceInterfaceLevel30.InitInfo info)
    {
        int rc = STAFResult.Ok;

        try
        {
            fHandle = new STAFHandle("STAF/SERVICE/" + info.name);
        }
        catch (STAFException e)
        {
            return new STAFResult(STAFResult.STAFRegistrationError,
                                  e.toString());
        }

        try
        {
            STAFResult res = new STAFResult();

            // Resolve the machine name variable for the local machine

            res = STAFUtil.resolveInitVar("{STAF/Config/Machine}", fHandle);

            if (res.rc != STAFResult.Ok) return res;

            fLocalMachineName = res.result;
            
            fServiceName = info.name;
            fEventServiceMachine = "local";
            fEventServiceName = "event";

            if (info.serviceJar != null)
            {
                // Create the Jython Library directory if it doesn't exist yet
                // from the jython library data in the service's jarfile.
                // Set the classpath to point to the service's jar file and to
                // the jython.jar.

                res = STAFServiceSharedJython.setupJython(
                    info.serviceJar, fHandle, info.writeLocation);

                if (res.rc != 0)
                {
                    System.out.println(res.result);
                    return res;
                }

                // Get the version of Jython distributed with the Cron service
                // Note the getJythonVersion method was added in STAF V3.0.0

                try
                {
                    fJythonVersion = STAFServiceSharedJython.getJythonVersion();
                }
                catch (Error err)
                {
                    return new STAFResult(
                        STAFResult.STAFRegistrationError,
                        fServiceName + " service V" + kVersion +
                        " requires STAF V" + kRequiredSTAFVersion +
                        " or later");
                }

                // Initialize the PythonInterpreter's python.home to point to
                // the shared jython installation.

                Properties p = new Properties();
                p.setProperty("python.home",
                              STAFServiceSharedJython.getJythonDirName());

                PythonInterpreter.initialize(System.getProperties(), p, null);
            }

            // Parse PARMS if provided in the STAX configuration line

            if (info.parms != null)
            {
                fParmsParser.addOption("EVENTSERVICEMACHINE", 1,
                                       STAFCommandParser.VALUEREQUIRED);
                fParmsParser.addOption("EVENTSERVICENAME", 1,
                                       STAFCommandParser.VALUEREQUIRED);
                res = handleParms(info);

                if (res.rc != STAFResult.Ok)
                {
                    return new STAFResult(
                        STAFResult.ServiceConfigurationError,
                        "Error validating parameters: RC=" + res.rc +
                        ", Result=" + res.result);
                }
            }

            // instantiate parsers as not case sensitive
            fRegisterParser = new STAFCommandParser(0, false);
            fListParser = new STAFCommandParser(0, false);
            fUnregisterParser = new STAFCommandParser(0, false);
            fVersionParser = new STAFCommandParser(0, false);

            fRegisterParser.addOption("REGISTER", 1,
                                      STAFCommandParser.VALUENOTALLOWED);

            fRegisterParser.addOption("MACHINE", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("PYTHONMACHINE", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOptionGroup("MACHINE PYTHONMACHINE", 1, 1);

            fRegisterParser.addOption("SERVICE", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("PYTHONSERVICE", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOptionGroup("SERVICE PYTHONSERVICE", 1, 1);

            fRegisterParser.addOption("REQUEST", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("PYTHONREQUEST", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOptionGroup("REQUEST PYTHONREQUEST", 1, 1);

            fRegisterParser.addOption("TYPE", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("SUBTYPE", 0,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("PREPARE", 0,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOptionNeed("REGISTER", "MACHINE PYTHONMACHINE");
            fRegisterParser.addOptionNeed("MACHINE PYTHONMACHINE", "REGISTER");

            fUnregisterParser.addOption("UNREGISTER", 1,
                                        STAFCommandParser.VALUENOTALLOWED);

            fUnregisterParser.addOption("ID", 1,
                                        STAFCommandParser.VALUEREQUIRED);

            fUnregisterParser.addOptionNeed("UNREGISTER", "ID");

            fListParser.addOption("LIST", 1,
                                  STAFCommandParser.VALUENOTALLOWED);

            fListParser.addOption("MACHINE", 1,
                                  STAFCommandParser.VALUEREQUIRED);

            fListParser.addOption("TYPE", 1,
                                  STAFCommandParser.VALUEREQUIRED);

            fListParser.addOption("LONG", 1,
                                  STAFCommandParser.VALUENOTALLOWED);

            fListParser.addOption("SETTINGS", 1, 
                                  STAFCommandParser.VALUENOTALLOWED);

            fListParser.addOptionGroup("MACHINE SETTINGS", 0, 1);

            fListParser.addOptionGroup("TYPE SETTINGS", 0, 1);

            fListParser.addOptionGroup("LONG SETTINGS", 0, 1);

            fVersionParser.addOption("VERSION", 1,
                                     STAFCommandParser.VALUENOTALLOWED);

            fVersionParser.addOption("JYTHON", 1,
                                     STAFCommandParser.VALUENOTALLOWED);

            fVersionParser.addOptionGroup("VERSION", 1, 1);
            fVersionParser.addOptionNeed("JYTHON", "VERSION");

            // Construct map class for a LIST request.

            fEMIDMapClass = new STAFMapClassDefinition(
                "STAF/Service/EventManager/EventManagerID");
            fEMIDMapClass.addKey("eventManagerID", "ID");
            fEMIDMapClass.addKey("machine", "Machine");
            fEMIDMapClass.addKey("service", "Service");
            fEMIDMapClass.addKey("request", "Request");
            fEMIDMapClass.addKey("type", "Event Type");
            fEMIDMapClass.addKey("subtype", "Event Subtype");

            // Construct map class for a LIST LONG request.

            fEventManagerIDMapClass = new STAFMapClassDefinition(
                "STAF/Service/EventManager/EventManagerID");
            fEventManagerIDMapClass.addKey("eventManagerID", "ID");
            fEventManagerIDMapClass.addKey("machine", "Machine");
            fEventManagerIDMapClass.addKey("machineType", "Machine Type");
            fEventManagerIDMapClass.addKey("service", "Service");
            fEventManagerIDMapClass.addKey("serviceType", "Service Type");
            fEventManagerIDMapClass.addKey("request", "Request");
            fEventManagerIDMapClass.addKey("requestType", "Request Type");
            fEventManagerIDMapClass.addKey("type", "Event Type");
            fEventManagerIDMapClass.addKey("subtype", "Event Subtype");
            fEventManagerIDMapClass.addKey("prepareScript", "Prepare Script");

            // Construct map-class for list settings information

            fSettingsMapClass = new STAFMapClassDefinition(
                "STAF/Service/EventManager/Settings");
            fSettingsMapClass.addKey(
                "eventServiceMachine", "Event Service Machine");
            fSettingsMapClass.addKey(
                "eventServiceName", "Event Service Name");

            // Resolve the line separator variable for the local machine

            res = STAFUtil.resolveInitVar("{STAF/Config/Sep/Line}", fHandle);

            if (res.rc != STAFResult.Ok) return res;

            fLineSep = res.result;
            
            // Resolve the file separator variable for the local machine

            res = STAFUtil.resolveInitVar("{STAF/Config/Sep/File}", fHandle);

            if (res.rc != STAFResult.Ok) return res;

            String fileSep = res.result;

            // Store data for the eventmanager service in directory:
            //   <STAF writeLocation>/service/<em service name (lower-case)>

            fHashtableFileDirectory = info.writeLocation;

            if (!fHashtableFileDirectory.endsWith(fileSep))
            {
                fHashtableFileDirectory += fileSep;
            }

            fHashtableFileDirectory = fHashtableFileDirectory +
                "service" + fileSep + fServiceName.toLowerCase();

            File dir = new File(fHashtableFileDirectory);

            if (!dir.exists())
            {
                dir.mkdirs();
            }

            fHashtableFileName = fHashtableFileDirectory + fileSep +
                "eventman.ser";

            loadHashtable();

            fQueueThread = new Thread(this);
            fQueueThread.start(); // this calls the run() method
        }
        catch (STAFException e)
        {
            rc = STAFResult.STAFRegistrationError;
            return new STAFResult(rc, e.toString());
        }

        return new STAFResult(rc);
    }

    private STAFResult handleParms(STAFServiceInterfaceLevel30.InitInfo info)
        throws STAFException
    {
        STAFCommandParseResult parseResult= fParmsParser.parse(info.parms);

        if (parseResult.rc != STAFResult.Ok)
        {
            return new STAFResult(
                STAFResult.InvalidRequestString, parseResult.errorBuffer);
        }

        if (parseResult.optionTimes("EVENTSERVICEMACHINE") > 0)
        {
            STAFResult res = STAFUtil.resolveInitVar(
                parseResult.optionValue("EVENTSERVICEMACHINE"), fHandle);

            if (res.rc != STAFResult.Ok) return res;

            fEventServiceMachine = res.result;
        }

        if (parseResult.optionTimes("EVENTSERVICENAME") > 0)
        {
            STAFResult res = STAFUtil.resolveInitVar(
                parseResult.optionValue("EVENTSERVICENAME"), fHandle);

            if (res.rc != STAFResult.Ok) return res;

            fEventServiceName = res.result;
        }

        return new STAFResult(STAFResult.Ok);
    }

    public STAFResult acceptRequest(STAFServiceInterfaceLevel30.RequestInfo
                                    info)
    {
        String lowerRequest = info.request.toLowerCase();

        // call the appropriate method to handle the command
        if (lowerRequest.startsWith("register"))
        {
            return handleRegister(info);
        }
        else if (lowerRequest.startsWith("unregister"))
        {
            return handleUnregister(info);
        }
        else if (lowerRequest.startsWith("list"))
        {
            return handleList(info);
        }
        else if (lowerRequest.startsWith("help"))
        {
            return handleHelp(info);
        }
        else if (lowerRequest.startsWith("version"))
        {
            return handleVersion(info);
        }
        else
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  "Unknown EventManager Request: " +
                                  lowerRequest);
        }
    }

    private STAFResult handleHelp(STAFServiceInterfaceLevel30.RequestInfo info)
    {
        // Verify the requester has at least trust level 1

        STAFResult trustResult = STAFUtil.validateTrust(
            1, fServiceName, "HELP", fLocalMachineName, info);

        if (trustResult.rc != STAFResult.Ok) return trustResult;

        // Return the help text for the service

        return new STAFResult(STAFResult.Ok,
         "EventManager Service Help" + fLineSep + fLineSep
         + "REGISTER   MACHINE <machine> | PYTHONMACHINE <machine>"
         + fLineSep +
         "           SERVICE <service> | PYTHONSERVICE <machine>"
         + fLineSep +
         "           REQUEST <request> | PYTHONREQUEST <request>"
         + fLineSep + "           TYPE <eventType> [SUBTYPE <eventSubType>]"
         + fLineSep + "           [PREPARE <script>]"
         + fLineSep + "UNREGISTER ID <registrationID>"
         + fLineSep + "LIST       <[MACHINE <machine>] [TYPE <eventType>] "
         + "[LONG]> | SETTINGS"
         + fLineSep + "VERSION    [JYTHON]" + fLineSep + "HELP");
    }

    private STAFResult handleList(STAFServiceInterfaceLevel30.RequestInfo info)
    {
        STAFCommandParseResult parsedRequest = fListParser.parse(info.request);

        if (parsedRequest.rc != STAFResult.Ok)
        {
            return new STAFResult(
                STAFResult.InvalidRequestString, parsedRequest.errorBuffer);
        }

        STAFMarshallingContext mc = new STAFMarshallingContext();
        List emIDList = new ArrayList();

        STAFResult sResult = new STAFResult(STAFResult.Ok, "");
        STAFResult resolvedResult = null;
        STAFResult resolvedValue = null;
        String machineParm = "";
        String typeParm = "";

        try
        {
            // Verify the requester has at least trust level 2

            STAFResult trustResult = STAFUtil.validateTrust(
                2, fServiceName, "LIST", fLocalMachineName, info);

            if (trustResult.rc != STAFResult.Ok) return trustResult;

            if (parsedRequest.optionTimes("SETTINGS") > 0)
            {
                // LIST SETTINGS

                mc.setMapClassDefinition(fSettingsMapClass);
                Map outputMap = fSettingsMapClass.createInstance();

                outputMap.put(
                    "eventServiceMachine", fEventServiceMachine);
                outputMap.put(
                    "eventServiceName", fEventServiceName);
                
                mc.setRootObject(outputMap);

                return new STAFResult(STAFResult.Ok, mc.marshall());
            }
            
            // Process the LIST registrationss request

            int registrationTableSize = fRegTable.size();

            if (registrationTableSize > 0)
            {
                // Resolve the MACHINE option

                resolvedValue = STAFUtil.resolveRequestVar(
                    parsedRequest.optionValue("machine"),
                    fHandle, info.requestNumber);
                
                if (resolvedValue.rc != 0) return resolvedValue;

                machineParm = resolvedValue.result;

                // Resolve the TYPE option

                resolvedValue = STAFUtil.resolveRequestVar(
                    parsedRequest.optionValue("type"),
                    fHandle, info.requestNumber);

                if (resolvedValue.rc != 0) return resolvedValue;

                typeParm = resolvedValue.result;

                // Check if the LONG option is specified

                boolean longFormat = false;

                if (parsedRequest.optionTimes("LONG") > 0)
                {
                    longFormat = true;
                    mc.setMapClassDefinition(fEventManagerIDMapClass);
                }
                else
                {
                    mc.setMapClassDefinition(fEMIDMapClass);
                }

                EventManagerData theData;
                boolean match;
                Enumeration registrationEnum = fRegTable.elements();
                Enumeration keys = fRegTable.keys();

                for (Enumeration e = fRegTable.elements();
                     e.hasMoreElements();)
                {
                    Integer id = (Integer) keys.nextElement();
                    match = false;
                    theData = (EventManagerData) e.nextElement();

                    if (machineParm.equals("") ||
                        machineParm.equalsIgnoreCase(theData.fMachine))
                    {
                        if (typeParm.equals("") ||
                            typeParm.equalsIgnoreCase(theData.fType))
                        {
                            match = true;
                        }
                    }

                    if (match)
                    {
                        // Create a map representing the matching registration
                        // and add it to the list of registrations

                        Map emIDMap;

                        if (longFormat)
                            emIDMap = fEventManagerIDMapClass.createInstance();
                        else
                            emIDMap = fEMIDMapClass.createInstance();

                        emIDMap.put("eventManagerID", id.toString());
                        emIDMap.put("machine", theData.fMachine);
                        emIDMap.put("service", theData.fService);
                        emIDMap.put("request", theData.fRequest);
                        emIDMap.put("type", theData.fType);
                        
                        if ((theData.fSubtype != null) &&
                            (!theData.fSubtype.equals("")))
                        {
                            emIDMap.put("subtype", theData.fSubtype);
                        }
                        
                        if (longFormat)
                        {
                            if (theData.fPythonMachine)
                                emIDMap.put("machineType", "Python");
                            else
                                emIDMap.put("machineType", "Literal");

                            if (theData.fPythonService)
                                emIDMap.put("serviceType", "Python");
                            else
                                emIDMap.put("serviceType", "Literal");

                            if (theData.fPythonRequest)
                                emIDMap.put("requestType", "Python");
                            else
                                emIDMap.put("requestType", "Literal");

                            if ((theData.fPrepare != null) &&
                                (!theData.fPrepare.equals("")))
                            {
                                emIDMap.put("prepareScript", theData.fPrepare);
                            }
                        }

                        emIDList.add(emIDMap);
                    }
                }
            }
        }
        catch (Exception e)
        {
            if (DEBUG)
            {
                e.printStackTrace();
            }

            return new STAFResult(STAFResult.JavaError,
                                  "Internal Java error.");
        }

        mc.setRootObject(emIDList);

        return new STAFResult(STAFResult.Ok, mc.marshall());
    }

    private STAFResult handleRegister(STAFServiceInterfaceLevel30.RequestInfo
                                      info)
    {
        STAFResult sResult = new STAFResult(STAFResult.Ok, "");
        String originMachine = info.endpoint;
        String machine = "";
        boolean pythonMachine = true;
        String service = "";
        boolean pythonService = true;
        String request = "";
        boolean pythonRequest = true;
        String type = "";
        String subtype = "";
        String prepare = "";

        STAFResult resolvedResult = null;
        STAFResult resolvedValue = null;
        STAFCommandParseResult parsedRequest =
            fRegisterParser.parse(info.request);

        if (parsedRequest.rc != STAFResult.Ok)
        {
            return new STAFResult(
                STAFResult.InvalidRequestString, parsedRequest.errorBuffer);
        }

        // For each option, if the user has specified a value, set the
        // corresponding local variable to the user-specified value.

        try
        {
            // Verify the requester has at least trust level 5

            STAFResult trustResult = STAFUtil.validateTrust(
                5, fServiceName, "REGISTER", fLocalMachineName, info);

            if (trustResult.rc != STAFResult.Ok) return trustResult;

            // Process the request

            if (parsedRequest.optionTimes("machine") > 0)
            {
                // Resolve the MACHINE option

                resolvedValue = STAFUtil.resolveRequestVar(
                    parsedRequest.optionValue("machine"),
                    fHandle, info.requestNumber);
                
                if (resolvedValue.rc != 0) return resolvedValue;

                machine = resolvedValue.result;
                pythonMachine = false;
            }
            else if (parsedRequest.optionTimes("pythonmachine") > 0)
            {
                machine = parsedRequest.optionValue("pythonmachine");
                pythonMachine = true;
            }

            if (parsedRequest.optionTimes("service") > 0)
            {
                // Resolve the SERVICE option

                resolvedValue = STAFUtil.resolveRequestVar(
                    parsedRequest.optionValue("service"),
                    fHandle, info.requestNumber);
                
                if (resolvedValue.rc != 0) return resolvedValue;

                service = resolvedValue.result;
                pythonService = false;
            }
            else if (parsedRequest.optionTimes("pythonservice") > 0)
            {
                service = parsedRequest.optionValue("pythonservice");
                pythonService = true;
            }

            if (parsedRequest.optionTimes("request") > 0)
            {
                // Resolve the REQUEST option

                resolvedValue = STAFUtil.resolveRequestVar(
                    parsedRequest.optionValue("request"),
                    fHandle, info.requestNumber);

                if (resolvedValue.rc != 0) return resolvedValue;

                request = resolvedValue.result;
                pythonRequest = false;
            }
            else if (parsedRequest.optionTimes("pythonrequest") > 0)
            {
                request = parsedRequest.optionValue("pythonrequest");
                pythonRequest = true;
            }

            // Resolve the TYPE option

            resolvedValue = STAFUtil.resolveRequestVar(
                parsedRequest.optionValue("type"), fHandle, info.requestNumber);

            if (resolvedValue.rc != 0) return resolvedValue;

            type = resolvedValue.result;

            if (parsedRequest.optionTimes("subtype") > 0)
            {
                // Resolve the SUBTYPE option

                resolvedValue = STAFUtil.resolveRequestVar(
                    parsedRequest.optionValue("subtype"),
                    fHandle, info.requestNumber);

                if (resolvedValue.rc != 0) return resolvedValue;
                
                subtype = resolvedValue.result;
            }

            if (parsedRequest.optionTimes("prepare") > 0)
            {
                prepare = parsedRequest.optionValue("prepare");
            }

            String registerRequest = "REGISTER TYPE " + type;

            if (!(subtype.equals("")))
            {
                registerRequest += " SUBTYPE " + subtype;
            }

            // Register with the Event service

            STAFResult registerResult = fHandle.submit2(
                fEventServiceMachine, fEventServiceName, registerRequest);

            if (registerResult.rc != 0)
            {
                return registerResult;
            }
            else
            {
                Integer index = new Integer(fEventManagerID++);

                fRegTable.put(
                    index, (Object) new EventManagerData(
                        originMachine, machine, pythonMachine, service,
                        pythonService, request, pythonRequest,
                        type, subtype, prepare));

                sResult = new STAFResult(STAFResult.Ok, index.toString());
                saveHashtable();

                String message = "Register ID=" + index +
                    " originMachine=" + originMachine +
                    " machine=" + machine + " pythonMachine=" + pythonMachine +
                    " service=" + service + " pythonService=" + pythonService +
                    " request=" + request + " pythonRequest=" + pythonRequest +
                    " prepare=" + prepare + " type=" + type +
                    " subtype=" + subtype;

                fHandle.submit2(
                    "local", "LOG", "LOG MACHINE LOGNAME " + fServiceName +
                    " LEVEL info MESSAGE " + STAFUtil.wrapData(message));
            }
        }
        catch (Exception e)
        {
            if (DEBUG)
            {
                e.printStackTrace();
            }

            return new STAFResult(STAFResult.JavaError,
                                  "Internal Java error.");
        }

        return sResult;
    }

    private STAFResult handleUnregister(STAFServiceInterfaceLevel30.RequestInfo
                                        info)
    {
        STAFResult sResult = new STAFResult(STAFResult.Ok, "");
        String service;
        String request;
        Integer id;

        STAFResult resolvedResult = null;
        STAFResult resolvedValue = null;
        STAFCommandParseResult parsedRequest = fUnregisterParser.parse(
            info.request);

        if (parsedRequest.rc != STAFResult.Ok)
        {
            return new STAFResult(
                STAFResult.InvalidRequestString, parsedRequest.errorBuffer);
        }

        // For each option, if the user has specified a value, set the
        // corresponding local variable to the user-specified value.

        try
        {
            // Verify the requester has at least trust level 4

            STAFResult trustResult = STAFUtil.validateTrust(
                4, fServiceName, "UNREGISTER", fLocalMachineName, info);

            if (trustResult.rc != STAFResult.Ok) return trustResult;

            // Resolve the ID option

            resolvedValue = STAFUtil.resolveRequestVarAndCheckInt(
                "id", parsedRequest.optionValue("id"),
                fHandle, info.requestNumber);

            if (resolvedValue.rc != 0) return resolvedValue;
            
            id = new Integer(resolvedValue.result);
            
            // Check if registration table contains the ID

            if (!fRegTable.containsKey(id))
            {
                return new STAFResult(
                    STAFResult.DoesNotExist, "ID " + id + " not found");
            }
            else
            {
                EventManagerData data = (EventManagerData)fRegTable.get(id);

                fRegTable.remove(id);

                if (fRegTable.isEmpty())
                {
                    fEventManagerID = 1;
                }

                saveHashtable();

                String message = "Unregister ID=" + id;

                fHandle.submit2(
                    "local", "LOG", "LOG MACHINE LOGNAME " + fServiceName +
                    " LEVEL info MESSAGE " + STAFUtil.wrapData(message));

                String unregType = data.fType;
                String unregSubtype = data.fSubtype;
                boolean unregWithEventService = true;

                // Only unregister with the Event Service if there are no
                // other fRegTable entries with this type and subtype

                for (Enumeration e = fRegTable.elements();
                    e.hasMoreElements();)
                {
                     EventManagerData theData =
                        (EventManagerData) e.nextElement();

                    if (unregType.equalsIgnoreCase(theData.fType) &&
                        theData.fSubtype.equalsIgnoreCase(unregSubtype))
                    {
                        unregWithEventService = false;
                    }
                }

                if (unregWithEventService)
                {
                    String unRegisterRequest = "UNREGISTER TYPE " + unregType;

                    if (!(data.fSubtype.equals("")))
                    {
                        unRegisterRequest += " SUBTYPE " + unregSubtype;
                    }

                        // Unregister with the Event service
                    STAFResult unRegisterResult = fHandle.submit2(
                        fEventServiceMachine, fEventServiceName,
                        unRegisterRequest);

                    if (unRegisterResult.rc != 0)
                    {
                        return unRegisterResult;
                    }
                }
            }
        }
        catch (Exception e)
        {
            if (DEBUG)
            {
                e.printStackTrace();
            }

            return new STAFResult(STAFResult.JavaError, "Internal Java error.");
        }

        return sResult;
    }

    private STAFResult handleVersion(
        STAFServiceInterfaceLevel30.RequestInfo info)
    {
        // Verify the requester has at least trust level 1

        STAFResult trustResult = STAFUtil.validateTrust(
            1, fServiceName, "VERSION", fLocalMachineName, info);

        if (trustResult.rc != STAFResult.Ok) return trustResult;

        // Parse the request

        STAFCommandParseResult parsedRequest = fVersionParser.parse(
            info.request);

        if (parsedRequest.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parsedRequest.errorBuffer);
        }

        if (parsedRequest.optionTimes("JYTHON") == 0)
        {
            // Return the version of the service
            return new STAFResult(STAFResult.Ok, kVersion);
        }
        else
        {
            // Return the version of Python packaged with the service
            return new STAFResult(STAFResult.Ok, fJythonVersion);
        }
    }


    public void run()
    {
        STAFResult queueGetResult;

        for (;;)
        {
            queueGetResult = fHandle.submit2("local", "QUEUE", "GET WAIT");

            if (queueGetResult.rc != 0)
            {
                // XXX: Do anything?
                continue;
            }

            STAFQueueMessage queueMessage = new STAFQueueMessage(
                queueGetResult.result);

            String queueType = queueMessage.type;

            if (queueType.equalsIgnoreCase(sQueueTypeEnd))
            {
                // This type of queue message indicates that the service
                // is terminating

                try
                {
                    fHandle.unRegister();
                }
                catch (STAFException ex)
                {
                    if (DEBUG)
                    {
                        ex.printStackTrace();
                    }
                }

                return;
            }
            else if (queueType.equalsIgnoreCase("STAF/RequestComplete"))
            {
                // A STAF/RequestComplete message is a map containing keys:
                //   requestNumber, rc, result

                Map messageMap = (Map)queueMessage.message;
                
                String reqNum = (String)messageMap.get("requestNumber");
                String rc = (String)messageMap.get("rc");
                Object result = (Object)messageMap.get("result");

                // Log a message about the completed request

                String level = "";

                if (rc.equals("0"))
                    level = "info";
                else
                    level = "error";
                
                String request = (String)fSubmittedRequests.get(reqNum);

                fHandle.submit2(
                    "local", "LOG", "LOG MACHINE LOGNAME " + fServiceName +
                    " LEVEL " + level + " MESSAGE " +
                    STAFUtil.wrapData("Request #" + reqNum + ": RC=" +
                                      rc + ", Result=" + result.toString() +
                                      ", Request submitted: " + request));

                // handle process service start requests

                String lowerRequest = request.toLowerCase();

                if ((lowerRequest.indexOf("service=process") > -1) &&
                    (lowerRequest.indexOf("request=start") > -1)  &&
                    (lowerRequest.indexOf("wait") == -1))
                {
                    String machine = queueMessage.machine;

                    String handle = result.toString();

                    STAFResult notifyRegResult;

                    notifyRegResult = fHandle.submit2(machine, "process",
                        "notify register onendofhandle " + handle);

                    fHandle.submit2(
                        "local", "log", "log machine logname " + fServiceName +
                        " level " + level + " message " +
                        STAFUtil.wrapData("Register for process completion" +
                                          " notification RC=" +
                                          notifyRegResult.rc + " Result=" +
                                          notifyRegResult.result + " Request=" +
                                          machine + " process" +
                                          " notify register onendofhandle " +
                                          handle));

                    if (notifyRegResult.rc ==
                        STAFResult.ProcessAlreadyComplete)
                    {
                        STAFResult freeHandleResult;

                        freeHandleResult = fHandle.submit2(machine, "process",
                            "free handle " + handle);

                        fHandle.submit2(
                            "local", "log", "log machine logname "
                            + fServiceName + " level " + level + " message " +
                            STAFUtil.wrapData("Free handle " + "RC=" +
                                              freeHandleResult.rc + " Result=" +
                                              freeHandleResult.result +
                                              " Request=" + machine +
                                              " process free handle " + handle)
                                              );;
                    }
                }

                // Remove the request from the submitted request map

                fSubmittedRequests.remove(reqNum);
            }
            else if (queueType.equalsIgnoreCase("STAF/PROCESS/END"))
            {
                Map messageMap = (Map)queueMessage.message;
                String handle = (String)messageMap.get("handle");

                String machine = queueMessage.machine;

                STAFResult freeHandleResult;

                freeHandleResult = fHandle.submit2(machine, "process",
                    "free handle " + handle);

                fHandle.submit2(
                    "local", "log", "log machine logname " + fServiceName +
                    " level info" + " message " +
                    STAFUtil.wrapData("Free handle " + "RC=" +
                                      freeHandleResult.rc + " Result=" +
                                      freeHandleResult.result + " Request=" +
                                      machine + " process free handle " +
                                      handle));;
            }
            else if (queueType.equalsIgnoreCase("STAF/Service/Event"))
            {
                if (!(queueMessage.message instanceof Map))
                {
                    System.out.println(
                        "EventManagerService: Unsupported message format.  " +
                        "Ignoring this message.\n" +
                        "priority=" + queueMessage.priority +
                        ", timestamp=" + queueMessage.timestamp +
                        ", machine=" + queueMessage.machine +
                        ", handleName=" + queueMessage.handleName +
                        "\nqueueType=" + queueType +
                        "\nmessage=" + queueMessage.message.toString());

                    // Ignore this message
                    continue;
                }

                Map messageMap = (Map)queueMessage.message;

                // create new Python Interpreter for each Event received

                fPythonInterpreter = new PythonInterpreter();

                Hashtable eventMap = new Hashtable();

                String eventService = (String)messageMap.get("eventServiceName");
                String eventID = (String)messageMap.get("eventID");
                String generatingMachine = (String)messageMap.get("machine");
                String generatingProcess = (String)messageMap.get("handleName");
                String generatingHandle = (String)messageMap.get("handle");
                String eventTimestamp = (String)messageMap.get("timestamp");
                String eventType = (String)messageMap.get("type");
                String eventSubtype = (String)messageMap.get("subtype");
                
                pySetVar("eventservice", eventService);
                eventMap.put(Py.java2py("eventservice"),
                             Py.java2py(eventService));

                pySetVar("eventid", eventID);
                eventMap.put(Py.java2py("eventid"), Py.java2py(eventID));

                pySetVar("generatingmachine", generatingMachine);
                eventMap.put(Py.java2py("generatingmachine"),
                             Py.java2py(generatingMachine));

                pySetVar("generatingprocess", generatingProcess);
                eventMap.put(Py.java2py("generatingprocess"),
                             Py.java2py(generatingProcess));

                pySetVar("generatinghandle", generatingHandle);
                eventMap.put(Py.java2py("generatinghandle"),
                             Py.java2py(generatingHandle));

                pySetVar("eventtimestamp", eventTimestamp);
                eventMap.put(Py.java2py("eventtimestamp"), 
                             Py.java2py(eventTimestamp));

                pySetVar("eventtype", eventType);
                eventMap.put(Py.java2py("eventtype"),
                             Py.java2py(eventType));

                pySetVar("eventsubtype", eventSubtype);
                eventMap.put(Py.java2py("eventsubtype"),
                             Py.java2py(eventSubtype));
                
                Map eventProperties = (Map)messageMap.get("propertyMap");
                Iterator iter = eventProperties.keySet().iterator();

                while (iter.hasNext())
                {
                    String name = (String)iter.next();
                    String value = (String)eventProperties.get(name);

                    pySetVar(name, value);
                    eventMap.put(Py.java2py(name), Py.java2py(value));
                }

                PyDictionary args = new PyDictionary(eventMap);
                pySetVar("eventinfo", args);

                EventManagerData theData;

                for (Enumeration e = fRegTable.keys();
                     e.hasMoreElements();)
                {
                    Integer id = (Integer)e.nextElement();

                    theData = (EventManagerData) fRegTable.get(id);

                    if (eventType.equalsIgnoreCase(theData.fType))
                    {
                        if (theData.fSubtype.equals("") ||
                            eventSubtype.equalsIgnoreCase(theData.fSubtype))
                        {
                            pySetVar("STAFEventManagerSubmit", "true");

                            String prepare = theData.fPrepare;
                            pyExec(prepare);

                            Object submit =
                                pyGetVar("STAFEventManagerSubmit");

                            if ((submit != null) &&
                                (submit.toString().equals("true")))
                            {
                                boolean pythonMachine = theData.fPythonMachine;
                                boolean pythonService = theData.fPythonService;
                                boolean pythonRequest = theData.fPythonRequest;

                                String machine = "";
                                String service = "";
                                String request = "";

                                if (pythonMachine)
                                {
                                    try
                                    {
                                        machine =
                                            pyStringEval(theData.fMachine);
                                    }
                                    catch (PyException ex)
                                    {
                                        String msg = "Error: " +
                                            "pythonmachine value is not valid"
                                            + " python code: " +
                                            theData.fMachine + " exception=" +
                                            ex;

                                        fHandle.submit2("local", "log",
                                            "log machine logname " +
                                            fServiceName +
                                            " level error message " +
                                            STAFUtil.wrapData(msg));

                                        continue;
                                    }
                                }
                                else
                                {
                                    machine = theData.fMachine;
                                }

                                if (pythonService)
                                {
                                    try
                                    {
                                        service  =
                                            pyStringEval(theData.fService);
                                    }
                                    catch (PyException ex)
                                    {
                                        String msg = "Error: " +
                                            "pythonservice value is not valid"
                                            + " python code: " +
                                            theData.fService + " exception=" +
                                            ex;

                                        fHandle.submit2("local", "log",
                                            "log machine logname " +
                                            fServiceName +
                                            " level error message " +
                                            STAFUtil.wrapData(msg));

                                        continue;
                                    }
                                }
                                else
                                {
                                    service = theData.fService;
                                }

                                if (pythonRequest)
                                {
                                    try
                                    {
                                        request =
                                             pyStringEval(theData.fRequest);
                                    }
                                    catch (PyException ex)
                                    {
                                        String msg = "Error: " +
                                            "pythonrequest value is not valid"
                                            + " python code: " +
                                            theData.fRequest + " exception=" +
                                            ex;

                                        fHandle.submit2("local", "log",
                                            "log machine logname " +
                                            fServiceName +
                                            " level error message " +
                                            STAFUtil.wrapData(msg));

                                         continue;
                                    }
                                    request = pyStringEval(theData.fRequest);
                                }
                                else
                                {
                                    request = theData.fRequest;
                                }

                                STAFResult matchResult = fHandle.submit2(
                                    STAFHandle.ReqQueue,
                                    machine, service, request);

                                fSubmittedRequests.put(matchResult.result,
                                    "machine=" + machine + " service=" +
                                    service + " request=" + request);

                                String message = "Submitted request #" +
                                    matchResult.result + " ID=" + id.toString()
                                    + " type=" + eventType +
                                    " subtype=" + eventSubtype +
                                    " prepare=" + theData.fPrepare +
                                    " eventservice=" + eventService +
                                    " eventid=" + eventID +
                                    " generatingmachine=" + generatingMachine +
                                    " generatingprocess=" + generatingProcess +
                                    " generatinghandle=" + generatingHandle +
                                    " eventtimestamp=" + eventTimestamp +
                                    " properties=" + eventProperties +
                                    " machine=" + machine + " service=" +
                                    service + " request=" + request;

                                fHandle.submit2("local", "log",
                                    "log machine logname " +
                                    fServiceName +
                                    " level info message " +
                                    STAFUtil.wrapData(message));
                            }
                        }
                    }
                }
            }
        }
    }

    public void loadHashtable()
    {
        try
        {
            File fHashtableFile = new File(fHashtableFileName);

            if (!fHashtableFile.exists())
            {
                fRegTable = new Hashtable();
                return;
            }
            else
            {
                ObjectInputStream ois =
                    new ObjectInputStream(new FileInputStream(fHashtableFile));
                fRegTable = (Hashtable)(ois.readObject());
                ois.close();

                for (Enumeration e = fRegTable.elements();
                     e.hasMoreElements();)
                {
                    EventManagerData data = (EventManagerData) e.nextElement();
                    STAFResult registerResult = fHandle.submit2(
                        fEventServiceMachine, fEventServiceName,
                        "REGISTER TYPE " + data.fType +
                        " SUBTYPE " + data.fSubtype);

                    if (registerResult.rc != 0)
                    {
                        fHandle.submit2(
                            "local", "LOG", "LOG MACHINE LOGNAME "
                            + fServiceName +
                            " LEVEL error MESSAGE " +
                            STAFUtil.wrapData("Error registering type " +
                            data.fType + " subtype " +
                            data.fSubtype + " with the "
                            + fEventServiceName + " service"));
                    }
                }

                fEventManagerID = findNextEventManagerID();

            }
        }
        catch (IOException e)
        {
            fRegTable = new Hashtable();
            return;
        }
        catch (ClassNotFoundException e)
        {
            fRegTable = new Hashtable();
            return;
        }
    }

    public void saveHashtable()
    {
        try
        {
            ObjectOutputStream oos =
                new ObjectOutputStream(new
                    FileOutputStream(fHashtableFileName));
            oos.writeObject(((Object)(fRegTable)));
            oos.close();
        }
        catch (IOException ex)
        {
            ex.printStackTrace();
        }
    }

    public int findNextEventManagerID()
    {
        int highestID = 0;
        for (Enumeration e = fRegTable.keys(); e.hasMoreElements();)
        {
            Integer id = (Integer) e.nextElement();
            if (id.intValue() > highestID)
            {
                highestID = id.intValue();
            }
        }

        return highestID + 1;
    }

    public STAFResult term()
    {
        fHandle.submit2(
            STAFHandle.ReqFireAndForget,
            "local", "QUEUE", "QUEUE TYPE " + sQueueTypeEnd +
            " MESSAGE " + STAFUtil.wrapData(""));

        // Un-register the service handle

        try
        {
            fHandle.unRegister();
        }
        catch (STAFException ex)
        {
            return new STAFResult(STAFResult.STAFRegistrationError,
                                  ex.toString());
        }

        return new STAFResult(STAFResult.Ok);
    }

    public String pyStringEval(String value) throws PyException
    {
        synchronized(fPythonInterpreter)
        {
            PyObject result = null;

            result = fPythonInterpreter.eval(value);

            if (result == null)
            {
                return null;
            }
            else
            {
                return result.toString();
            }
        }
    }

    public void pySetVar(String varName, Object value)
    {
        synchronized(fPythonInterpreter)
        {
            try
            {
                fPythonInterpreter.set(varName, value);
            }
            catch (PyException e)
            {
                fHandle.submit2(
                    "local", "LOG", "LOG MACHINE LOGNAME " +
                    fServiceName + " LEVEL error MESSAGE " +
                    STAFUtil.wrapData("PySetVar PyException caught setting " +
                                      " var=" + varName + " to value=" +
                                      value + e.toString()));
            }
        }
    }

    public Object pyGetVar(String varName)
    {
        synchronized(fPythonInterpreter)
        {
            try
            {
                return fPythonInterpreter.get(varName);
            }
            catch (PyException e)
            {
                fHandle.submit2(
                    "local", "LOG", "LOG MACHINE LOGNAME " +
                    fServiceName + " level error message " +
                    STAFUtil.wrapData("PyGetVar PyException caught getting " +
                                      " var=" + varName + " to value=" +
                                      varName + e.toString()));

                return null;
            }
        }
    }

    public void pyExec(String value)
    {
        synchronized(fPythonInterpreter)
        {
            try
            {
                fPythonInterpreter.exec(value);
            }
            catch (PyException e)
            {
                e.printStackTrace();
            }
        }
    }

}
