/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

package com.ibm.staf.service.stax;
import org.w3c.dom.Node;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.NodeList;

public class STAXMessageActionFactory implements STAXActionFactory
{
    static final String STAX_MESSAGE = new String("Message");

    private static String fDTDInfo =
"\n" +
"<!--================= The Message Element ========================== -->\n" +
"<!--\n" +
"     Generates an event and makes the message value available to the\n" +
"     STAX Job Monitor.  The message must evaluate via Python to a string.\n\n" +
"     The log attribute is evaluated via Python to a boolean.  If it\n" +
"     evaluates to true, the message text will also be logged in the STAX\n" +
"     Job User log.  The log attribute defaults to the STAXLogMessage\n" +
"     variable whose value defaults to 0 (false) but can by changed within\n" +
"     the STAX job to turn on logging.\n\n" +
"     The log level is ignored if the log attribute does not evaluate to\n" +
"     true.  It defaults to 'info'.  If specified, it must evaluate via\n" +
"     Python to a string containing one of the following STAF Log Service\n" +
"     logging levels:\n" +
"       fatal, warning, info, trace, trace2, trace3, debug, debug2,\n" +
"       debug3, start, stop, pass, fail, status, user1, user2, user3,\n" +
"       user4, user5, user6, user7, user8\n\n" +
"     If an if attribute is specified and it evaluates via Python to\n" +
"     false, the message element is ignored.\n" +
"-->\n" +
"<!ELEMENT message     (#PCDATA)>\n" +
"<!ATTLIST message\n" +
"          log         CDATA       \"STAXLogMessage\"\n" +
"          level       CDATA       \"'info'\"\n" +
"          if          CDATA       \"1\"\n" +
">\n";

    public String getDTDInfo()
    {
        return fDTDInfo;
    }

    public String getDTDTaskName()
    {
        return "message";
    }

    public STAXAction parseAction(STAX staxService, STAXJob job,
                                  org.w3c.dom.Node root) throws STAXException
    {
        STAXMessageAction action = new STAXMessageAction();

        NamedNodeMap attrs = root.getAttributes();

        for (int i = 0; i < attrs.getLength(); ++i)
        {
            Node thisAttr = attrs.item(i);

            String errorInfo = "\n  Element: " + root.getNodeName() +
                               "  Attribute: " + thisAttr.getNodeName();

            if (thisAttr.getNodeName().equals("log"))
            {
                action.setLogAttr(STAXUtil.parseAndCompileForPython(
                    thisAttr.getNodeValue(), errorInfo));
            }
            else if (thisAttr.getNodeName().equals("level"))
            {
                if (thisAttr.getNodeValue().equals(""))
                {
                    throw new STAXInvalidXMLAttributeException(
                        errorInfo + "\nThe value for the level attribute " +
                        "cannot be blank");
                }

                action.setLevel(STAXUtil.parseAndCompileForPython(
                    thisAttr.getNodeValue(), errorInfo));
            }
            else if (thisAttr.getNodeName().equals("if"))
            {
                action.setIf(STAXUtil.parseAndCompileForPython(
                    thisAttr.getNodeValue(), errorInfo));
            }
            else
            {
                throw new STAXInvalidXMLAttributeException(
                    root.getNodeName() + ": " + thisAttr.getNodeName());
            }
        }

        String errorInfo = "\n  Element: " + root.getNodeName();

        NodeList children = root.getChildNodes();

        for (int i = 0; i < children.getLength(); ++i)
        {
            Node thisChild = children.item(i);

            if (thisChild.getNodeType() == Node.COMMENT_NODE)
            {
                /* Do nothing */
            }
            else if (thisChild.getNodeType() == Node.TEXT_NODE)
            {
                action.setMessageValue(STAXUtil.parseAndCompileForPython(
                    thisChild.getNodeValue(), errorInfo));
            }
            else
            {
                throw new STAXInvalidXMLNodeTypeException(
                    Integer.toString(thisChild.getNodeType()));
            }
        }

        return action;
    }
}
