/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2003                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAFString.h"
#include "STAF_iostream.h"
#include <stdlib.h>


void submitRequest(const STAFHandlePtr handlePtr,
                   const STAFString &machine,
                   const STAFString &service,
                   const STAFString &request,
                   bool showResult)
{
    STAFResultPtr result = handlePtr->submit(machine, service, request);

    if (result->rc != 0)
    {
        cout << endl << "STAF " << machine << " " << service << " "
             << request << endl
             << "Error submitting request, RC: " << result->rc << endl;

        if (result->result.length() != 0)
            cout << "Additional info: " << result->result << endl;
    }
    else
    {
        cout << endl << "STAF " << machine << " " << service << " "
             << request << " worked" << endl;
        if (showResult)
        {
            cout << STAFObject::unmarshall(result->result)->asFormattedString() << endl;
        }
    }
}

int main(int argc, char **argv)
{
    if (argc != 4 && argc != 5 && argc != 6)
    {
        cout << "Usage: " << endl;
        cout << " " << argv[0] << " <TargetMachine> <UserId> <Password> [<Auth> [<ToMachine>]]" << endl;
        cout << "Examples: " << endl;
        cout << " " << argv[0] << " staf2c User1 Password1 (if SampleAuth is default auth)" << endl;
        cout << " " << argv[0] << " local  User1 Password1 SampleAuth" << endl;
        cout << " " << argv[0] << " staf2c User1 Password1 SampleAuth staf3a" << endl;
        return 1;
    }
    
    STAFString machine(argv[1]);  // Can be local
    STAFString userId(argv[2]);
    STAFString pw(argv[3]);
    
    cout << endl;

    STAFHandlePtr handlePtr;
    unsigned int rc = 0;
    STAFString handleName = "STAF/Test/Authentication";
    
    rc = STAFHandle::create(handleName, handlePtr);

    if (rc != 0)
    {
        cout << "Error registering with STAF, RC: " << rc << endl;
        return rc;
    }
    
    // Assign the orgMachine name by resolving {STAF/Config/Machine}

    STAFString orgMachine;
    
    STAFString local("local");
    STAFString service("VAR");
    STAFString request("RESOLVE STRING {STAF/Config/Machine}");

    STAFResultPtr result = handlePtr->submit(local, service, request);

    if (result->rc != 0)
    {
        cout << endl << "STAF " << local << " " << service << " "
             << request << endl
             << "Error submitting request, RC: " << result->rc << endl;

        return result->rc;
    }
    else
    {
        orgMachine = result->result;
        cout << "OrgMachine   = " << orgMachine << endl;
    }

    STAFString authenticator;

    if (argc > 4)
    {
        authenticator = argv[4];
        cout << "Authenticator= " << authenticator << endl;
    }
    else
    {
        // Get the local system's default authenticator by resolving {STAF/Config/DefaultAuthenticator}
        
        STAFString local("local");
        STAFString service("VAR");
        STAFString request("RESOLVE STRING {STAF/Config/DefaultAuthenticator}");

        STAFResultPtr result = handlePtr->submit(local, service, request);

        if (result->rc != 0)
        {
            cout << endl << "STAF " << local << " " << service << " "
                 << request << endl
                 << "Error submitting request, RC: " << result->rc << endl;

            return result->rc;
        }
        else
        {
            authenticator = result->result;
            cout << "Authenticator= " << authenticator << endl;
        }
    }
    
    cout << "UserId       = " << userId << endl;
    cout << "Password     = " << pw << endl;
    cout << "TargetMachine= " << machine << endl;

    // Get version of Org Machine.  If >= 3.0, list its authenticators

    service = "MISC";
    request = "VERSION";

    result = handlePtr->submit(orgMachine, service, request);

    if (result->rc != 0)
    {
        cout << endl << "STAF " << orgMachine << " " << service << " "
             << request << endl
             << "Error submitting request, RC: " << result->rc << endl;

        if (result->result.length() != 0)
            cout << "Additional info: " << result->result << endl;
    }
    else
    {
        cout << "OrgMachineVersion= " << result->result << endl;

        if (result->result >= "3.0.0")
        {
            // List its authenticators

            service = "SERVICE";
            request = "LIST AUTHENTICATORS";

            result = handlePtr->submit(orgMachine, service, request);

            if (result->rc != 0)
            {
                cout << endl << "STAF " << orgMachine << " " << service << " "
                     << request << endl
                     << "Error submitting request, RC: " << result->rc << endl;

                if (result->result.length() != 0)
                    cout << "Additional info: " << result->result << endl;
            }
            else
            {
                cout << "OrgMachineAuthenticators:" << endl << endl
                     << STAFObject::unmarshall(result->result)->asFormattedString() << endl;
            }
        }
    }
    
    // List all trusts for Org Machine.

    service = "TRUST";
    request = "LIST";

    result = handlePtr->submit(orgMachine, service, request);

    if (result->rc != 0)
    {
        cout << endl << "STAF " << orgMachine << " " << service << " "
             << request << endl
             << "Error submitting request, RC: " << result->rc << endl;

        if (result->result.length() != 0)
            cout << "Additional info: " << result->result << endl;
    }
    else
    {
        cout << "OrgMachineTrusts:" << endl << endl
             << STAFObject::unmarshall(result->result)->asFormattedString() << endl;
    }
    
    // Get version of Target Machine.  If >= 3.0, list its authenticators

    service = "MISC";
    request = "VERSION";

    result = handlePtr->submit(machine, service, request);

    if (result->rc != 0)
    {
        cout << endl << "STAF " << machine << " " << service << " "
             << request << endl
             << "Error submitting request, RC: " << result->rc << endl;

        if (result->result.length() != 0)
            cout << "Additional info: " << result->result << endl;
    }
    else
    {
        cout << "TargetMachineVersion= " << result->result << endl;

        if (result->result >= "3.0.0")
        {
            // List its authenticators

            service = "SERVICE";
            request = "LIST AUTHENTICATORS";

            result = handlePtr->submit(machine, service, request);

            if (result->rc != 0)
            {
                cout << endl << "STAF " << machine << " " << service << " "
                     << request << endl
                     << "Error submitting request, RC: " << result->rc << endl;

                if (result->result.length() != 0)
                    cout << "Additional info: " << result->result << endl;
            }
            else
            {
                cout << "TargetMachineAuthenticators:" << endl << endl
                     << STAFObject::unmarshall(result->result)->asFormattedString() << endl;
            }
        }
    }
    
    // List all trusts for Target Machine.

    service = "TRUST";
    request = "LIST";

    result = handlePtr->submit(machine, service, request);

    if (result->rc != 0)
    {
        cout << endl << "STAF " << machine << " " << service << " "
             << request << endl
             << "Error submitting request, RC: " << result->rc << endl;

        if (result->result.length() != 0)
            cout << "Additional info: " << result->result << endl;
    }
    else
    {
        cout << "TargetMachineTrusts:" << endl << endl
             << STAFObject::unmarshall(result->result)->asFormattedString() << endl;
    }

    STAFString toMachine;
    if (argc == 6)
      toMachine = argv[5];
    else
      toMachine = orgMachine;

    cout << "ToMachine    = " << toMachine << endl;

    if (toMachine != orgMachine)
    {
        // Get version of To Machine.  If >= 3.0, list its authenticators

        service = "MISC";
        request = "VERSION";

        result = handlePtr->submit(toMachine, service, request);

        if (result->rc != 0)
        {
            cout << endl << "STAF " << toMachine << " " << service << " "
                 << request << endl
                 << "Error submitting request, RC: " << result->rc << endl;

            if (result->result.length() != 0)
                cout << "Additional info: " << result->result << endl;
        }
        else
        {
            cout << "ToMachineVersion = " << result->result << endl;

            if (result->result >= "3.0.0")
            {
                // List its authenticators

                service = "SERVICE";
                request = "LIST AUTHENTICATORS";

                result = handlePtr->submit(toMachine, service, request);

                if (result->rc != 0)
                {
                    cout << endl << "STAF " << toMachine << " " << service << " "
                         << request << endl
                         << "Error submitting request, RC: " << result->rc << endl;

                    if (result->result.length() != 0)
                        cout << "Additional info: " << result->result << endl;
                }
                else
                {
                    cout << "ToMachineAuthenticators:" << endl << endl
                         << STAFObject::unmarshall(result->result)->asFormattedString() << endl;
                }
            }
        }

        // List all trusts for To Machine.

        service = "TRUST";
        request = "LIST";

        result = handlePtr->submit(toMachine, service, request);

        if (result->rc != 0)
        {
            cout << endl << "STAF " << toMachine << " " << service << " "
                 << request << endl
                 << "Error submitting request, RC: " << result->rc << endl;

            if (result->result.length() != 0)
                cout << "Additional info: " << result->result << endl;
        }
        else
        {
            cout << "ToMachineTrusts:" << endl << endl
                 << STAFObject::unmarshall(result->result)->asFormattedString() << endl;
        }
    }


    // Part 1: Authenticate the handle with a valid userid/password with
    //         trust level 5.  Verify can access all internal services.

    cout << endl << "Part 1: Authenticate the handle with a valid userid/password with"
         << endl << "        trust level 5.  Verify can access all internal services"
         << endl;

    service = "HANDLE";
    request = "AUTHENTICATE USER " + userId + " CREDENTIALS " + pw;

    if (authenticator.length() > 0)
      request += " AUTHENTICATOR " + authenticator;

    submitRequest(handlePtr, local, service, request, false);

    service = "HANDLE";
    request = "LIST HANDLES NAME " + handleName;
    submitRequest(handlePtr, local, service, request, true);

    service = "VAR";
    request = "RESOLVE STRING {STAF/Handle/User}";
    submitRequest(handlePtr, local, service, request, true);

    service = "MISC";
    request = "WHOAMI";
    submitRequest(handlePtr, local, service, request, true);

    service = "MISC";
    request = "WHOAMI";
    submitRequest(handlePtr, machine, service, request, true);

    service = "DELAY";
    request = "DELAY 1000";
    submitRequest(handlePtr, machine, service, request, false);

    service = "PROCESS";
    request = "START SHELL COMMAND :13:java -version WAIT";
    submitRequest(handlePtr, machine, service, request, false);

    service = "FS";
    request = "COPY FILE {STAF/Config/ConfigFile} TOFILE C:/test.xxx";
    submitRequest(handlePtr, machine, service, request, false);

    request = "COPY DIRECTORY {STAF/Config/STAFRoot}/docs TODIRECTORY C:/temp/docs";
    submitRequest(handlePtr, machine, service, request, false);
    
    request = "COPY FILE {STAF/Config/ConfigFile} TOFILE C:/test.xxx TOMACHINE " + toMachine;
    submitRequest(handlePtr, machine, service, request, false);
    
    request = "COPY DIRECTORY {STAF/Config/STAFRoot}/docs TODIRECTORY C:/temp/docs TOMACHINE " + toMachine;
    submitRequest(handlePtr, machine, service, request, false);
    
    request = "COPY FILE {STAF/Config/ConfigFile} TOFILE C:/test.xxx TOMACHINE " + machine;
    submitRequest(handlePtr, local, service, request, false);
    
    request = "COPY DIRECTORY {STAF/Config/STAFRoot}/docs TODIRECTORY C:/temp/docs TOMACHINE " + machine;
    submitRequest(handlePtr, local, service, request, false);

    service = "LOG";
    request = "SET NORESOLVEMESSAGE";   // Requires trust level 5
    submitRequest(handlePtr, machine, service, request, true);
    
    // Part 2: Authenticate the handle with a different userid
    
    cout << endl << "Part2: Authenticate the handle with a different userid/pw" << endl;

    userId = "User2";
    pw = "Password2";

    service = "HANDLE";
    request = "AUTHENTICATE USER " + userId + " CREDENTIALS " + pw;
    if (authenticator.length() > 0)
      request += " AUTHENTICATOR " + authenticator;

    cout << endl << "Try to re-authenticate an already authenticated handle" << endl;

    result = handlePtr->submit(local, service, request);

    if (result->rc == kSTAFHandleAlreadyAuthenticated)
    {
        cout << endl << "Re-authentication of the handle failed as expected" << endl;
    }
    else
    {
        cout << endl << "ERROR: Re-authentication of the already "
             << "authenticated handle worked " << endl;
    }
    
    cout << endl << "Un-authenticate the handle" << endl;

    request = "UNAUTHENTICATE";
    submitRequest(handlePtr, local, service, request, false);

    request = "AUTHENTICATE USER " + userId + " CREDENTIALS " + pw;
   
    if (authenticator.length() > 0)
      request += " AUTHENTICATOR " + authenticator;

    submitRequest(handlePtr, local, service, request, false);
    
    service = "HANDLE";
    request = "LIST HANDLES NAME " + handleName;
    submitRequest(handlePtr, local, service, request, true);
    
    service = "VAR";
    request = "RESOLVE STRING {STAF/Handle/User}";
    submitRequest(handlePtr, local, service, request, true);

    service = "DELAY";
    request = "DELAY 1000";
    submitRequest(handlePtr, machine, service, request, false);

    service = "PROCESS";
    request = "START SHELL COMMAND :13:java -version WAIT";
    submitRequest(handlePtr, machine, service, request, false);

    service = "FS";
    request = "COPY FILE {STAF/Config/ConfigFile} TOFILE C:/test.xxx";
    submitRequest(handlePtr, machine, service, request, false);
    
    request = "COPY DIRECTORY {STAF/Config/STAFRoot}/docs TODIRECTORY C:/temp/docs";
    submitRequest(handlePtr, machine, service, request, false);

    request = "COPY FILE {STAF/Config/ConfigFile} TOFILE C:/test.xxx TOMACHINE " + toMachine;
    submitRequest(handlePtr, machine, service, request, false);

    request = "COPY DIRECTORY {STAF/Config/STAFRoot}/docs TODIRECTORY C:/temp/docs TOMACHINE " + toMachine;
    submitRequest(handlePtr, machine, service, request, false);

    request = "COPY FILE {STAF/Config/ConfigFile} TOFILE C:/test.xxx TOMACHINE " + machine;
    submitRequest(handlePtr, local, service, request, false);
    
    request = "COPY DIRECTORY {STAF/Config/STAFRoot}/docs TODIRECTORY C:/temp/docs TOMACHINE " + machine;
    submitRequest(handlePtr, local, service, request, false);

    // Part 3: Try to authenticate the handle with a userid and wrong password
    
    cout << endl << "Part3: Authenticate the handle with a userid and wrong password" << endl;

    cout << endl << "Un-authenticate the handle" << endl;

    service = "HANDLE";
    request = "UNAUTHENTICATE";
    submitRequest(handlePtr, local, service, request, false);

    userId = "User2";
    pw = "PasswordXYZ";

    service = "HANDLE";
    request = "AUTHENTICATE USER " + userId + " CREDENTIALS " + pw;
    if (authenticator.length() > 0)
      request += " AUTHENTICATOR " + authenticator;

    submitRequest(handlePtr, local, service, request, false);
    
    service = "HANDLE";
    request = "LIST HANDLES NAME " + handleName;
    submitRequest(handlePtr, local, service, request, true);
    
    service = "VAR";
    request = "RESOLVE STRING {STAF/Handle/User}";
    submitRequest(handlePtr, local, service, request, true);

    service = "DELAY";
    request = "DELAY 1000";
    submitRequest(handlePtr, machine, service, request, false);

    service = "FS";
    request = "COPY FILE {STAF/Config/ConfigFile} TOFILE C:/test.xxx";
    submitRequest(handlePtr, machine, service, request, false);

    request = "COPY DIRECTORY {STAF/Config/STAFRoot}/docs TODIRECTORY C:/temp/docs";
    submitRequest(handlePtr, machine, service, request, false);
    
    request = "COPY FILE {STAF/Config/ConfigFile} TOFILE C:/test.xxx TOMACHINE " + toMachine;
    submitRequest(handlePtr, machine, service, request, false);

    request = "COPY DIRECTORY {STAF/Config/STAFRoot}/docs TODIRECTORY C:/temp/docs TOMACHINE " + toMachine;
    submitRequest(handlePtr, machine, service, request, false);
    
    request = "COPY FILE {STAF/Config/ConfigFile} TOFILE C:/test.xxx TOMACHINE " + machine;
    submitRequest(handlePtr, local, service, request, false);
    
    request = "COPY DIRECTORY {STAF/Config/STAFRoot}/docs TODIRECTORY C:/temp/docs TOMACHINE " + machine;
    submitRequest(handlePtr, local, service, request, false);

    // Part 4: Authenticate the handle with a userid that has trust level 0
    
    cout << endl << "Part4: Authenticate the handle with a userid that has trust level 0" << endl;

    cout << endl << "Un-authenticate the handle" << endl;

    service = "HANDLE";
    request = "UNAUTHENTICATE";
    submitRequest(handlePtr, local, service, request, false);

    userId = "User5";
    pw = "Password5";

    service = "HANDLE";
    request = "AUTHENTICATE USER " + userId + " CREDENTIALS " + pw;
    if (authenticator.length() > 0)
      request += " AUTHENTICATOR " + authenticator;

    submitRequest(handlePtr, local, service, request, false);
    
    service = "HANDLE";
    request = "LIST HANDLES NAME " + handleName;
    submitRequest(handlePtr, local, service, request, true);

    service = "VAR";
    request = "RESOLVE STRING {STAF/Handle/User}";
    submitRequest(handlePtr, local, service, request, true);

    service = "DELAY";
    request = "DELAY 1000";
    submitRequest(handlePtr, machine, service, request, false);

    service = "FS";
    request = "COPY FILE {STAF/Config/ConfigFile} TOFILE C:/test.xxx";
    submitRequest(handlePtr, machine, service, request, false);

    request = "COPY DIRECTORY {STAF/Config/STAFRoot}/docs TODIRECTORY C:/temp/docs";
    submitRequest(handlePtr, machine, service, request, false);
    
    request = "COPY FILE {STAF/Config/ConfigFile} TOFILE C:/test.xxx TOMACHINE " + toMachine;
    submitRequest(handlePtr, machine, service, request, false);
    
    request = "COPY DIRECTORY {STAF/Config/STAFRoot}/docs TODIRECTORY C:/temp/docs TOMACHINE " + toMachine;
    submitRequest(handlePtr, machine, service, request, false);
    
    request = "COPY FILE {STAF/Config/ConfigFile} TOFILE C:/test.xxx TOMACHINE " + machine;
    submitRequest(handlePtr, local, service, request, false);
    
    request = "COPY DIRECTORY {STAF/Config/STAFRoot}/docs TODIRECTORY C:/temp/docs TOMACHINE " + machine;
    submitRequest(handlePtr, local, service, request, false);

    return 0;
}
